"""
Network related helper code

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""
import logging
import platform
import subprocess
from privex.helpers.exceptions import BoundaryException, NetworkUnreachable
from privex.helpers import plugin
from ipaddress import ip_address, IPv4Address, IPv6Address
from typing import Union

log = logging.getLogger(__name__)

__all__ = [
    'ip_to_rdns', '_check_boundaries', 'ip4_to_rdns', 'ip6_to_rdns', 'ip_is_v4', 'ip_is_v6',
    'ping', 'BoundaryException', 'NetworkUnreachable'
]

try:
    from dns.resolver import Resolver, NoAnswer, NXDOMAIN
    
    def asn_to_name(as_number: Union[int, str], quiet: bool = True) -> str:
        """
        Look up an integer Autonomous System Number and return the human readable
        name of the organization.

        Usage:

        >>> asn_to_name(210083)
        'PRIVEX, SE'
        >>> asn_to_name('13335')
        'CLOUDFLARENET - Cloudflare, Inc., US'

        This helper function requires ``dnspython>=1.16.0``, it will not be visible unless
        you install the dnspython package in your virtualenv, or systemwide::
        
            pip3 install dnspython
        

        :param int/str as_number: The AS number as a string or integer, e.g. 210083 or '210083'
        :param bool quiet:        (default True) If True, returns 'Unknown ASN' if a lookup fails. 
                                  If False, raises a KeyError if no results are found.
        :raises KeyError:         Raised when a lookup returns no results, and ``quiet`` is set to False.
        :return str as_name:      The name and country code of the ASN, e.g. 'PRIVEX, SE'
        """
        
        try:
            res = Resolver().query('AS{}.asn.cymru.com'.format(as_number), "TXT")
            if len(res) > 0:
                # res[0] is formatted like such: "15169 | US | arin | 2000-03-30 | GOOGLE - Google LLC, US" with
                # literal quotes. we need to strip them, split by pipe, extract the last element, then strip spaces.
                asname = str(res[0]).strip('"').split('|')[-1:][0].strip()
                return str(asname)
            raise NoAnswer('privex.helpers.net.asn_to_name returned no results.')
        except (NoAnswer, NXDOMAIN):
            if quiet:
                return 'Unknown ASN'
            raise KeyError('ASN {} was not found, or server did not respond.'.format(as_number))
    
    __all__ += ['asn_to_name']
    plugin.HAS_DNSPYTHON = True
    
except ImportError:
    log.debug('privex.helpers.net failed to import "dns.resolver" (pypi package "dnspython"), skipping some helpers')
    pass


def ip_to_rdns(ip: str, boundary: bool = False, v6_boundary: int = 32, v4_boundary: int = 24) -> str:
    """
    Converts an IPv4 or IPv6 address into an in-addr domain

    Default boundaries: IPv4 - 24 bits, IPv6 - 32 bits

    **Examples:**

        >>> ip_to_rdns('127.0.0.1') # IPv4 to arpa format
            '1.0.0.127.in-addr.arpa'

        >>> ip_to_rdns('2001:dead:beef::1') # IPv6 to arpa format
            '1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.f.e.e.b.d.a.e.d.1.0.0.2.ip6.arpa'

        >>> ip_to_rdns('2001:dead:beef::1', boundary=True) # IPv6 32-bit boundary to arpa
            'd.a.e.d.1.0.0.2.ip6.arpa'

    :param str ip: IPv4 or IPv6 address
    :param bool boundary: If True, return the base (boundary) domain to place NS/SOA
    :param int v6_boundary: Bits for IPv6 boundary. Must be dividable by 4 bits (nibble)
    :param int v4_boundary: Bits for IPv4 boundary. Must be dividable by 8 bits (octet)

    :raises ValueError: When IP address is invalid
    :raises BoundaryException: When boundary for IPv4/v6 is invalid

    :return str rdns_domain: in-addr.arpa format, e.g. ``0.0.127.in-addr.arpa``
    :return str rdns_domain: ip6.arpa format, e.g. ``0.8.e.f.ip6.arpa``
    """
    v6_boundary = int(v6_boundary)
    v4_boundary = int(v4_boundary)
    # sanity check
    if boundary:
        # will raise a BoundaryException if v4/v6 is invalid
        _check_boundaries(v4_boundary, v6_boundary)
    # ip_address helps us detect the type, and uncompress v6 addresses
    ip_obj = ip_address(ip)

    # IPv6 Addresses use nibbles and ip6.arpa
    ip_type = type(ip_obj)
    if ip_type == IPv6Address:
        return ip6_to_rdns(ip_obj, v6_boundary, boundary)
    elif ip_type == IPv4Address:
        return ip4_to_rdns(ip_obj, v4_boundary, boundary)
    else:
        raise ValueError("Not a v4 nor v6 address...? Type was: {}".format(ip_type))


def _check_boundaries(v4_boundary: int, v6_boundary: int):
    """
    Boundary validator for :py:func:`.ip_to_rdns`

    :raises BoundaryException: When either the v4 or v6 boundary is invalid
    """
    if v6_boundary > 128 or v6_boundary < 4:
        raise BoundaryException('v6_boundary must be at least 4 bits, and less than 128 (was: {})'.format(v6_boundary))
    if v4_boundary > 32 or v4_boundary < 8:
        raise BoundaryException('v4_boundary must be at least 8 bits, and less than 32 (was: {})'.format(v4_boundary))
    if v6_boundary % 4 != 0:
        raise BoundaryException('v6_boundary not dividable by 4 bits (was: {})'.format(v6_boundary))
    if v4_boundary % 8 != 0:
        raise BoundaryException('v4_boundary not dividable by 8 bits (was: {})'.format(v4_boundary))


def ip4_to_rdns(ip_obj: IPv4Address, v4_boundary: int = 24, boundary: bool = False) -> str:
    """
    Internal function for getting the rDNS domain for a given v4 address. Use :py:func:`.ip_to_rdns` unless 
    you have a specific need for this one.

    :param IPv4Address ip_obj: An IPv4 ip_address() object to get the rDNS domain for
    :param int v4_boundary: 8-32 bits. If ``boundary`` is True, return the base rDNS domain at this boundary.
    :param bool boundary: If True, cut off the rDNS domain to the given ``v4_boundary``
    :return str rdns_domain: in-addr.arpa format, e.g. ``0.0.127.in-addr.arpa``
    """
    addr = ip_obj.reverse_pointer
    if boundary:
        exploded = ip_obj.exploded
        # each octet of ipv4 is 8 bits. only take the first x octets for boundary
        octets = int(v4_boundary / 8)
        addr_bounded = exploded.split('.')[:octets][::-1]  # array of octets, trim octets, and reverse the array
        addr_joined = '.'.join(addr_bounded)  # glue the array back together
        addr = addr_joined + '.in-addr.arpa'  # and finally add the arpa subdomain
    return addr


def ip6_to_rdns(ip_obj: IPv6Address, v6_boundary: int = 32, boundary: bool = False) -> str:
    """
    Internal function for getting the rDNS domain for a given v6 address. Use :py:func:`.ip_to_rdns` unless 
    you have a specific need for this one.

    :param IPv6Address ip_obj: An IPv4 ip_address() object to get the rDNS domain for
    :param int v6_boundary: 8-128 bits. If ``boundary`` is True, return the base rDNS domain at this boundary.
    :param bool boundary: If True, cut off the rDNS domain to the given ``v6_boundary``
    :return str rdns_domain: ip6.arpa format, e.g. ``0.8.e.f.ip6.arpa``
    """
    # first uncompress the IP
    expanded = ip_obj.exploded
    # Remove the colons
    # if we're not doing boundary, this will be used later, otherwise it'll just be replaced anyway
    addr = expanded.replace(':', '')
    if boundary:
        # if we need a boundary address, use v6_boundary to strip off the excess characters
        del_chars = int(v6_boundary / 4)        # 4 bits = 1 character (nibble)
        addr = addr[:del_chars]                 # now we have the boundary portion, e.g. 200100fa
    addr_reverse = addr[::-1]                   # reverse the address, e.g. af001002
    addr_joined = '.'.join(list(addr_reverse))  # join together with dots, e.g. a.f.0.0.1.0.0.2
    return addr_joined + '.ip6.arpa'            # and finally, return the completed string, a.f.0.0.1.0.0.2.ip6.arpa


def ip_is_v4(ip: str) -> bool:
    """
    Determines whether an IP address is IPv4 or not

    :param str ip: An IP address as a string, e.g. 192.168.1.1
    :raises ValueError: When the given IP address ``ip`` is invalid
    :return bool: True if IPv6, False if not (i.e. probably IPv4)
    """
    return type(ip_address(ip)) == IPv4Address


def ip_is_v6(ip: str) -> bool:
    """
    Determines whether an IP address is IPv6 or not

    :param str ip: An IP address as a string, e.g. 192.168.1.1
    :raises ValueError: When the given IP address ``ip`` is invalid
    :return bool: True if IPv6, False if not (i.e. probably IPv4)
    """
    return type(ip_address(ip)) == IPv6Address


def ping(ip: str, timeout: int = 30) -> bool:
    """
    Sends a ping to a given IPv4 / IPv6 address. Tested with IPv4+IPv6 using ``iputils-ping`` on Linux, as well as the
    default IPv4 ``ping`` utility on Mac OSX (Mojave, 10.14.6).
    
    Fully supported when using Linux with the ``iputils-ping`` package. Only IPv4 support on Mac OSX.
    
    **Example Usage**::
    
        >>> from privex.helpers import ping
        >>> if ping('127.0.0.1', 5) and ping('::1', 10):
        ...     print('Both 127.0.0.1 and ::1 are up')
        ... else:
        ...     print('127.0.0.1 or ::1 failed to respond to a ping within the given timeout.')
    
    **Known Incompatibilities**:
    
     * NOT compatible with IPv6 addresses on OSX due to the lack of a timeout argument with ``ping6``
     * NOT compatible with IPv6 addresses when using ``inetutils-ping`` on Linux due to separate ``ping6`` command

    :param str ip: An IP address as a string, e.g. ``192.168.1.1`` or ``2a07:e00::1``
    :param int timeout: (Default: 30) Number of seconds to wait for a response from the ping before timing out
    :raises ValueError: When the given IP address ``ip`` is invalid or ``timeout`` < 1
    :return bool: ``True`` if ping got a response from the given IP, ``False`` if not
    """
    ip_obj = ip_address(ip)   # verify IP is valid (this will throw if it isn't)
    if timeout < 1:
        raise ValueError('timeout value cannot be less than 1 second')
    opts4 = {
        'Linux': ["/bin/ping", "-c1", f"-w{timeout}"],
        'Darwin': ["/sbin/ping", "-c1", f"-t{timeout}"]
    }
    opts6 = {'Linux':  ["/bin/ping", "-c1", f"-w{timeout}"]}
    opts = opts4 if ip_is_v4(ip_obj) else opts6
    if platform.system() not in opts:
        raise NotImplementedError(f"{__name__}.ping is not fully supported on platform '{platform.system()}'...")
    
    with subprocess.Popen(opts[platform.system()] + [ip], stdout=subprocess.PIPE, stderr=subprocess.PIPE) as proc:
        out, err = proc.communicate()
        err = err.decode('utf-8')
        if 'network is unreachable' in err.lower():
            raise NetworkUnreachable(f'Got error from ping: "{err}"')
        
        return 'bytes from {}'.format(ip) in out.decode('utf-8')



"""
Functions, classes and/or types which either **are**, or are related to Python
variable storage types (``dict``, ``tuple``, ``list``, ``set`` etc.)

Object-like Dictionaries (dict's)
---------------------------------

Have you ever wanted a dictionary that works like an object, where you can get/set dictionary keys using 
attributes (``x.something``) as easily as you can with items (``x['something']``)?

We did. So we invented :class:`.DictObject`, a sub-class of the built-in :class:`dict`, making it compatible
with most functions/methods which expect a :class:`dict` (e.g. :meth:`json.dumps`).

You can create a new :class:`.DictObject` and use it just like a ``dict``, or you can convert an existing
``dict`` into a ``DictObject`` much like you'd cast any other builtin type.

It can also easily be cast back into a standard ``dict`` when needed, without losing any data.

Creating a new DictObject and using it
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Since :class:`.DictObject` is a subclass of the builtin :class:`dict`, you can instantiate a new
:class:`.DictObject` in the same way you would use the standard :class:`dict` class::

        >>> d = DictObject(hello='world')
        >>> d
        {'hello': 'world'}
        >>> d['hello']
        'world'
        >>> d.hello
        'world'
        >>> d.lorem = 'ipsum'
        >>> d['orange'] = 'banana'
        >>> d
        {'hello': 'world', 'lorem': 'ipsum', 'orange': 'banana'}


Converting an existing dictionary (dict) into a DictObject
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You can convert an existing ``dict`` into a :class:`.DictObject` in the same way you'd convert
any other object into a ``dict``::

    >>> y = {"hello": "world", "example": 123}
    >>> x = DictObject(y)
    >>> x.example
    123
    >>> x['hello']
    'world'
    >>> x.hello = 'replaced'
    >>> x
    {'hello': 'replaced', 'example': 123}

It also works vice versa, you can convert a :class:`.DictObject` instance back into a :class:`dict` just as
easily as you converted the `dict` into a `DictObject`.

    >>> z = dict(x)
    >>> z
    {'hello': 'replaced', 'example': 123}



Dict-able NamedTuple's
----------------------

While :func:`collections.namedtuple`'s can be useful, they have some quirks, such as not being able to access
fields by item/key (``x['something']``). They also expose a method ``._asdict()``, but cannot be directly casted
into a :class:`dict` using ``dict(x)``.

Our :func:`.dictable_namedtuple` collection is designed to fix these quirks.

What is a dictable_namedtuple and why use it?
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Unlike the normal :func:`.namedtuple` types, ``dictable_namedtuple``s add extra convenience functionality:

    * Can access fields via item/key: ``john['first_name']``

    * Can convert instance into a dict simply by casting: ``dict(john)``

    * Can set new items/attributes on an instance, even if they weren't previously defined.
    
    * NOTE: You cannot edit an original namedtuple field defined on the type, those remain read only

There are three functions available for working with ``dictable_namedtuple`` classes/instances,
each for different purposes.

  * :py:func:`.dictable_namedtuple` - Create a new ``dictable_namedtuple`` type for instantiation.

  * :py:func:`.convert_dictable_namedtuple` - Convert an existing **namedtuple instance** (not a type/class) into
    a ``dictable_namedtuple`` instance.
  
  * :py:func:`.subclass_dictable_namedtuple` - Convert an existing **namedtuple type/class** (not an instance) into
    a ``dictable_namedtuple`` type for instantiation.


Importing dictable_namedtuple functions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. code-block:: python

    from collections import namedtuple
    from privex.helpers import dictable_namedtuple, convert_dictable_namedtuple, subclass_dictable_namedtuple

Creating a NEW dictable_namedtuple type and instance
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you're **creating a new Named Tuple**, and you want it to support dictionary-like access, and
have it able to be converted into a dict simply through ``dict(my_namedtuple)``, then you want
:py:func:`.dictable_namedtuple`

.. code-block:: python

    Person = dictable_namedtuple('Person', 'first_name last_name')
    john = Person('John', 'Doe')
    dave = Person(first_name='Dave', last_name='Smith')
    print(dave['first_name'])       # Prints:  Dave
    print(dave.first_name)          # Prints:  Dave
    print(john[1])                  # Prints:  Doe
    print(dict(john))               # Prints:  {'first_name': 'John', 'last_name': 'Doe'}

Converting an existing namedtuple instance into a dictable_namedtuple instance
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


If you have **existing Named Tuple instances**, e.g. returned from a python library, then you can use
:py:func:`.convert_dictable_namedtuple` to convert them into ``dictable_namedtuple``'s and gain all the
functionality mentioned at the start of this section. 

.. code-block:: python

    Person = namedtuple('Person', 'first_name last_name')  # This is an existing namedtuple "type" or "class"
    john = Person('John', 'Doe')  # This is an existing namedtuple instance
    john.first_name               # This works on a standard namedtuple. Returns: John
    john[1]                       # This works on a standard namedtuple. Returns: Doe
    john['first_name']            # However, this would throw a TypeError.
    dict(john)                    # And this would throw a ValueError.

    # We can now convert 'john' into a dictable_namedtuple, which will retain the functionality of a
    # namedtuple, but add to the functionality by allowing dict-like key access, updating/creating new
    # fields, as well as painlessly casting to a dictionary.

    d_john = convert_dictable_namedtuple(john)
    d_john.first_name               # Returns: John
    d_john[1]                       # Returns: Doe
    d_john['first_name']            # Returns: 'John'
    dict(d_john)                    # Returns: {'first_name': 'John', 'last_name': 'Doe'}

Converting an existing namedtuple type/class into a dictable_namedtuple type/class
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you have **existing Named Tuple type/class** then you can use :py:func:`.subclass_dictable_namedtuple`
to convert the type/class into a ``dictable_namedtuple`` type/class and gain all the functionality mentioned 
at the start of this section. (**NOTE:** it's usually easier to just replace your ``namedtuple`` calls 
with ``dictable_namedtuple``)

.. code-block:: python

    Person = namedtuple('Person', 'first_name last_name')  # This is an existing namedtuple "type" or "class"
    # We can now convert the 'Person' type into a dictable_namedtuple type.
    d_Person = subclass_dictable_namedtuple(Person)
    # Then we can use this converted type to create instances of Person with dictable_namedtuple functionality.
    john = d_Person('John', 'Doe')
    john.first_name               # Returns: John
    john[1]                       # Returns: Doe
    john['first_name']            # Returns: 'John'
    dict(john)                    # Returns: {'first_name': 'John', 'last_name': 'Doe'}



**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import inspect
import sys
from collections import namedtuple, OrderedDict
from typing import Dict, Optional, NamedTuple, Union, Type
import logging

log = logging.getLogger(__name__)


class DictObject(dict):
    """
    A very simple :class:`dict` wrapper, which allows you to read and write dictionary keys using attributes
    (dot notation) PLUS standard item (key / square bracket notation) access.

    **Example Usage (creating and using a new DictObject)**::

        >>> d = DictObject(hello='world')
        >>> d
        {'hello': 'world'}
        >>> d['hello']
        'world'
        >>> d.hello
        'world'
        >>> d.lorem = 'ipsum'
        >>> d['orange'] = 'banana'
        >>> d
        {'hello': 'world', 'lorem': 'ipsum', 'orange': 'banana'}

    **Example Usage (converting an existing dict)**::

        >>> y = {"hello": "world", "example": 123}
        >>> x = DictObject(y)
        >>> x.example
        123
        >>> x['hello']
        'world'
        >>> x.hello = 'replaced'
        >>> x
        {'hello': 'replaced', 'example': 123}

    """

    def __getattr__(self, item):
        """When an attribute is requested, e.g. ``x.something``, forward it to ``dict['something']``"""
        if hasattr(super(), item):
            return super().__getattribute__(item)
        return super().__getitem__(item)

    def __setattr__(self, key, value):
        """When an attribute is set, e.g. ``x.something = 'abcd'``, forward it to ``dict['something'] = 'abcd'``"""
        if hasattr(super(), key):
            return super().__setattr__(key, value)
        return super().__setitem__(key, value)
    

class OrderedDictObject(OrderedDict):
    """
    Ordered version of :class:`.DictObject` - dictionary with attribute access.
    See :class:`.DictObject`
    """
    def __getattr__(self, item):
        """When an attribute is requested, e.g. ``x.something``, forward it to ``dict['something']``"""
        if hasattr(super(), item):
            return super().__getattribute__(item)
        return super().__getitem__(item)
    
    def __setattr__(self, key, value):
        """When an attribute is set, e.g. ``x.something = 'abcd'``, forward it to ``dict['something'] = 'abcd'``"""
        if hasattr(super(), key):
            return super().__setattr__(key, value)
        return super().__setitem__(key, value)


class MockDictObj(DictObject):
    """
    This is a masqueraded :class:`.DictObject` made to look like the builtin :class:`dict` by
    editing the class name, qualname and module.

    It may improve compatibility when passing :class:`.DictObject` to certain third-party 
    functions/methods.

    Note: this isn't enough to fool a ``type(x) is dict`` check.
    """


MockDictObj.__name__ = 'dict'
MockDictObj.__qualname__ = 'dict'
MockDictObj.__module__ = 'builtins'


def is_namedtuple(*objs) -> bool:
    """
    Takes one or more objects as positional arguments, and returns ``True`` if ALL passed objects
    are namedtuple instances

    **Example usage**

    First, create or obtain one or more NamedTuple objects::

        >>> from collections import namedtuple
        
        >>> Point, Person = namedtuple('Point', 'x y'), namedtuple('Person', 'first_name last_name')

        >>> pt1, pt2 = Point(1.0, 5.0), Point(2.5, 1.5)
        >>> john = Person('John', 'Doe')
    
    We'll also create a ``tuple``, ``dict``, and ``str`` to show they're detected as invalid::

        >>> normal_tuple, tst_dict, tst_str = (1, 2, 3,), dict(hello='world'), "hello world"
    
    First we'll call :func:`.is_namedtuple` with our Person NamedTuple object ``john``::

        >>> is_namedtuple(john)
        True
    
    As expected, the function shows ``john`` is in-fact a named tuple.

    Now let's try it with our two Point named tuple's ``pt1`` and ``pt2``, plus our Person named tuple ``john``.

        >>> is_namedtuple(pt1, john, pt2)
        True
    
    Since all three arguments were named tuples (even though pt1/pt2 and john are different types), the function
    returns ``True``.

    Now we'll test with a few objects that clearly aren't named tuple's::

        >>> is_namedtuple(tst_str)   # Strings aren't named tuples.
        False
        >>> is_namedtuple(normal_tuple)    # A plain bracket tuple is not a named tuple.
        False
        >>> is_namedtuple(john, tst_dict)  # ``john`` is a named tuple, but a dict isn't, thus False is returned.
        False


    Original source: https://stackoverflow.com/a/2166841

    :param Any objs: The objects (as positional args) to check whether they are a NamedTuple
    :return bool is_namedtuple: ``True`` if all passed ``objs`` are named tuples.
    """
    if len(objs) == 0: raise AttributeError("is_namedtuple expects at least one argument")

    for x in objs:
        t = type(x)
        b = t.__bases__
        if tuple not in b: return False
        f = getattr(t, '_fields', None)
        if not isinstance(f, tuple): return False
        if not all(type(n) == str for n in f): return False
    return True


def convert_dictable_namedtuple(nt_instance, typename=None, module=None, **kwargs) -> Union[NamedTuple, Dict]:
    """
    Convert an existing :func:`collections.namedtuple` instance into a dictable_namedtuple instance.

    **Example**

    First we create a namedtuple type ``Person``

        >>> from collections import namedtuple
        >>> Person = namedtuple('Person', 'first_name last_name')
    
    Next we create an instance of ``Person`` called John Doe, and we can confirm it's a normal namedtuple, as we
    can't access first_name by item/key.

        >>> john = Person('John', 'Doe')
        >>> john['first_name']
        TypeError: tuple indices must be integers or slices, not str
    
    Using :func:`.convert_dictable_namedtuple`, we can convert ``john`` from a normal ``namedtuple``, into
    a ``dictable_namedtuple``.
    
    This enables many convenience features (see :func:`.dictable_namedtuple` for more info) 
    such as easy casting to a :class:`dict`, and accessing fields by item/key (square brackets)::

        >>> from privex.helpers import convert_dictable_namedtuple
        >>> d_john = convert_dictable_namedtuple(john)
        >>> d_john
        Person(first_name='John', last_name='Doe')
        >>> d_john['first_name']
        'John'
        >>> dict(d_john)
        {'first_name': 'John', 'last_name': 'Doe'}
    

    :param nt_instance: An instantiated namedtuple object (using a type returned from :func:`collections.namedtuple`)
    :param str typename: Optionally, you can change the name of your instance's class, e.g. if you provide a ``Person`` 
                         instance, but you set this to ``Man``, then this will return a ``Man`` instance, like so:
                         ``Man(first_name='John', last_name='Doe')``
    :param str module: Optionally, you can change the module that the type class belongs to. Otherwise it will inherit the module path
                       from the class of your instance.
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple instance will not
                         allow new fields to be created via attribute / item setting.
    :return dictable_namedtuple: The instance you passed ``nt_instance``, converted into a dictable_namedtuple
    """

    nt_class = nt_instance.__class__
    module = nt_class.__module__ if module is None else module

    dnt_class = subclass_dictable_namedtuple(nt_class, typename=typename, module=module, **kwargs)
    return dnt_class(**nt_instance._asdict())


def subclass_dictable_namedtuple(named_type: type, typename=None, module=None, **kwargs) -> type:
    """
    Convert an existing :func:`collections.namedtuple` **type** into a dictable_namedtuple.

    If you have an INSTANCE of a type (e.g. it has data attached), use :func:`.convert_dictable_namedtuple`

    **Example**::

        >>> from collections import namedtuple
        >>> from privex.helpers import subclass_dictable_namedtuple
        >>> # Create a namedtuple type called 'Person'
        >>> orig_Person = namedtuple('Person', 'first_name last_name')
        >>> # Convert the 'Person' type into a dictable_namedtuple
        >>> Person = subclass_dictable_namedtuple(orig_Person)
        >>> john = Person('John', 'Doe')   # Create an instance of this dictable_namedtuple Person
        >>> john['middle_name'] = 'Davis'
    
    :param type named_type: A NamedTuple type returned from :func:`collections.namedtuple`
    :param str typename: Optionally, you can change the name of your type, e.g. if you provide a ``Person`` 
                         class type, but you set this to ``Man``, then this will return a ``Man`` class type.
    :param str module: Optionally, you can change the module that the type class belongs to. Otherwise it will
                       inherit the module path from ``named_type``.
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple type will not allow
                         new fields to be created via attribute / item setting.
    :return type dictable_namedtuple: Your ``named_type`` converted into a dictable_namedtuple type class.
    """
    typename = named_type.__name__ if typename is None else typename
    module = named_type.__module__ if module is None else module
    read_only = kwargs.pop('read_only', False)

    _dt = make_dict_tuple(typename, ' '.join(named_type._fields), read_only=read_only)

    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass

    if module is not None:
        _dt.__module__ = module
    return _dt


def make_dict_tuple(typename, field_names, *args, **kwargs):
    """
    Generates a :func:`collections.namedtuple` type, with added / modified methods injected to make it
    into a ``dictable_namedtuple``.
    
    Note: You probably want to be using :func:`.dictable_namedtuple` instead of calling this directly.
    """
    read_only = kwargs.pop('read_only', False)
    module = kwargs.pop('module', None)
    
    # Create a namedtuple type to use as a base
    BaseNT = namedtuple(typename, field_names, **kwargs)

    def __init__(self, *args, **kwargs):
        self.__dict__['_extra_items'] = dict()
        for i, a in enumerate(list(args)):
            self.__dict__[self._fields[i]] = a
        for k, a in kwargs.items():
            self.__dict__[k] = a
    
    def __iter__(self):
        """This ``__iter__`` method allows for casting a dictable_namedtuple instance using ``dict(my_nt)``"""
        for k in self._fields: yield (k, getattr(self, k),)

    def __getitem__(self, item):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt['abc']`` or ``my_nt[0]``"""
        if type(item) is int:
            return self.__dict__[self._fields[item]]
        return getattr(self, item)

    def __getattr__(self, item):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt.abcd``"""
        try:
            _v = object.__getattribute__(self, '_extra_items')
            return _v[item]
        except (KeyError, AttributeError):
            return object.__getattribute__(self, item)

    def __setitem__(self, key, value):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt['abc'] = 'def'``"""
        if hasattr(self, key):
            return tuple.__setattr__(self, key, value)
        if self._READ_ONLY:
            raise KeyError(f"{self.__class__.__name__} is read only. You cannot set a non-existent field.")
        self._extra_items[key] = value
        if key not in self._fields:
            tuple.__setattr__(self, '_fields', self._fields + (key,))

    def __setattr__(self, key, value):
        """Handles when a dictable_namedtuple instance is accessed like ``my_nt.abcd = 'def'``"""
        if key in ['_extra_items', '_fields'] or key in self._fields:
            return tuple.__setattr__(self, key, value)
        if self._READ_ONLY:
            raise AttributeError(f"{self.__class__.__name__} is read only. You cannot set a non-existent field.")
        self._extra_items[key] = value
        if key not in self._fields:
            tuple.__setattr__(self, '_fields', self._fields + (key,))

    def _asdict(self):
        """
        The original namedtuple ``_asdict`` doesn't work with our :meth:`.__iter__`, so we override it
        for compatibility. Simply calls ``return dict(self)`` to convert the instance to a dict.
        """
        return dict(self)

    def __repr__(self):
        _n = ', '.join(f"{name}='{getattr(self, name)}'" for name in self._fields)
        return f"{self.__class__.__name__}({_n})"

    # Inject our methods defined above into the namedtuple type BaseNT
    BaseNT.__getattr__ = __getattr__
    BaseNT.__getitem__ = __getitem__
    BaseNT.__setitem__ = __setitem__
    BaseNT.__setattr__ = __setattr__
    BaseNT._asdict = _asdict
    BaseNT.__repr__ = __repr__
    BaseNT.__iter__ = __iter__
    BaseNT.__init__ = __init__
    BaseNT._READ_ONLY = read_only
    
    # Create a class for BaseNT with tuple + object mixins, allowing things like __dict__ to function properly
    # and allowing for tuple.__setattr__ / object.__getattribute__ calls.
    class K(BaseNT, tuple, object):
        pass

    # Get the calling module so we can overwrite the module name of the class.
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass
    
    # Overwrite the type name + module to match the originally requested typename
    K.__name__ = BaseNT.__name__
    K.__qualname__ = BaseNT.__qualname__
    K.__module__ = module

    return K


def dictable_namedtuple(typename, field_names, *args, **kwargs) -> Union[Type[namedtuple], dict]:
    """
    Creates a dictable_namedtuple type for instantiation (same usage as :func:`collections.namedtuple`) - unlike
    namedtuple, dictable_namedtuple instances allow item (dict-like) field access, support writing and can be
    painlessly converted into dictionaries via ``dict(my_namedtuple)``.

    Named tuple instances created from ``dictable_namedtuple`` types are generally backwards compatible with
    any code that expects a standard :func:`collections.namedtuple` type instance.

    **Quickstart**

        >>> from privex.helpers import dictable_namedtuple
        >>> # Define a dictable_namedtuple type of 'Person', which has two fields - first_name and last_name
        >>> p = dictable_namedtuple('Person', 'first_name last_name')
        >>> john = p('John', 'Doe')    # Alternatively you can do p(first_name='John', last_name='Doe')
        >>> john.first_name            # You can retrieve keys either via attributes (dot notation)
        'John'
        >>> john['last_name']          # Via named keys (square brackets)
        'Doe'
        >>> john[1]                    # Or, via indexed keys (square brackets, with integer keys)
        'Doe'
        >>> john.middle_name = 'Davis' # You can also update / set new keys via attribute/key/index
        >>> dict(john)                 # Newly created keys will show up as normal in dict(your_object)
        {'first_name': 'John', 'last_name': 'Doe', 'middle_name': 'Davis'}
        >>> john                       # As well as in the representation in the REPL or when str() is called.
        Person(first_name='John', last_name='Doe', middle_name='Davis')

    This function adds / overrides the following methods on the generated namedtuple type:

    * _asdict
    * __iter__
    * __getitem__
    * __getattribute__
    * __setitem__
    * __setattr__
    * __repr__

    Extra functionality compared to the standard :func:`.namedtuple` generated classes:

        * Can access fields via item/key: ``john['first_name']``

        * Can convert instance into a dict simply by casting: ``dict(john)``

        * Can set new items/attributes on an instance, even if they weren't previously defined.
          ``john['middle_name'] = 'Davis'`` or ``john.middle_name = 'Davis'``

    **Example Usage**

    First we'll create a named tuple typle called ``Person``, which takes two arguments, first_name and last_name.

        >>> from privex.helpers import dictable_namedtuple
        >>> Person = dictable_namedtuple('Person', 'first_name last_name')
    
    Now we'll create an instance of ``Person`` called ``john``. These instances look like normal ``namedtuple``'s, and
    should be generally compatible with any functions/methods which deal with named tuple's.

        >>> john = Person('John', 'Doe')   # Alternatively you can do Person(first_name='John', last_name='Doe')
        >>> john
        Person(first_name='John', last_name='Doe')
    
    Unlike a normal ``namedtuple`` type instance, we can access fields by attribute (``.first_name``), index (``[0]``), 
    AND by item/key name (``['last_name']``).

        >>> john.first_name
        'John'
        >>> john[0]
        'John'
        >>> john['last_name']
        'Doe'
    
    Another potentially useful feature, is that you can also update / create new fields, via your preferred method
    of field notation (other than numbered indexes, since those don't include a field name)::

        >>> john['middle_name'] = 'Davis'
        >>> john.middle_name = 'Davis'

    We can also convert ``john`` into a standard dictionary, with a simple ``dict(john)`` cast. You can see that
    the new field we added (``middle_name``) is present in the dictionary serialized format.

        >>> dict(john)
        {'first_name': 'John', 'last_name': 'Doe', 'middle_name': 'Davis'}
    

    :param str typename: The name used for the namedtuple type/class
    :param str field_names: One or more field names separated by spaces, e.g. ``'id first_name last_name address'``
    :key bool read_only: (Default: ``False``) If set to ``True``, the outputted dictable_namedtuple instance will not
                         allow new fields to be created via attribute / item setting.
    :return Type[namedtuple] dict_namedtuple: A dict_namedtuple type/class which can be instantiated with the given
                             ``field_names`` via positional or keyword args.
    """
    module = kwargs.get('module', None)
    read_only = kwargs.pop('read_only', False)
    
    # As per namedtuple's comment block, we need to set __module__ to the frame
    # where the named tuple is created, otherwise it can't be pickled properly.
    # This also ensures that __module__ would match that of a normal namedtuple()
    if module is None:
        try:
            module = sys._getframe(1).f_globals.get('__name__', '__main__')
        except (AttributeError, ValueError):
            pass

    return make_dict_tuple(typename, field_names, module=module, read_only=read_only, *args, **kwargs)


class Dictable:
    """
    A small abstract class for use with Python 3.7 dataclasses.

    Allows dataclasses to be converted into a ``dict`` using the standard ``dict()`` function:

        >>> @dataclass
        >>> class SomeData(Dictable):
        ...     a: str
        ...     b: int
        ...
        >>> mydata = SomeData(a='test', b=2)
        >>> dict(mydata)
        {'a': 'test', 'b': 2}

    Also allows creating dataclasses from arbitrary dictionaries, while ignoring any extraneous dict keys.

    If you create a dataclass using a ``dict`` and you have keys in your ``dict`` that don't exist in the dataclass,
    it'll generally throw an error due to non-existent kwargs:

        >>> mydict = dict(a='test', b=2, c='hello')
        >>> sd = SomeData(**mydict)
        TypeError: __init__() got an unexpected keyword argument 'c'

    Using ``from_dict`` you can simply trim off any extraneous dict keys:

        >>> sd = SomeData.from_dict(**mydict)
        >>> sd.a, sd.b
        ('test', 2)
        >>> sd.c
        AttributeError: 'SomeData' object has no attribute 'c'



    """

    def __iter__(self):
        # Allow casting into dict()
        for k, v in self.__dict__.items(): yield (k, v,)

    @classmethod
    def from_dict(cls, env):
        # noinspection PyArgumentList
        return cls(**{
            k: v for k, v in env.items()
            if k in inspect.signature(cls).parameters
        })


class Mocker(object):
    """
    This mock class is designed to be used either to act as a stand-in "noop" (no operation) object, which
    could be used either as a drop-in replacement for a failed module / class import, or for certain unit tests.
    
    If you need additional functionality such as methods having actual behaviour, you can set attributes on a
    Mocker instance to either a lambda, or point them at a real function/method::
    
        >>> m = Mocker()
        >>> m.some_func = lambda a: a+1
        >>> m.some_func(5)
        6
    
    
    **Example use case - fallback for unimportant module imports**
    
    Below is a real world example of using :class:`.Mocker` and :py:func:`privex.helpers.decorators.mock_decorator`
    to simulate :py:mod:`pytest` - allowing your tests to run under the standard :py:mod:`unittest` framework if
    a user doesn't have pytest (as long as your tests aren't critically dependent on PyTest).
    
    Try importing ``pytest`` then fallback to a mock pytest::
    
        >>> try:
        ...     import pytest
        ... except ImportError:
        ...     from privex.helpers import Mocker, mock_decorator
        ...     print('Failed to import pytest. Using privex.helpers.Mocker to fake pytest.')
        ...     # Make pytest pretend to be the class 'module' (the class actually used for modules)
        ...     pytest = Mocker.make_mock_class('module')
        ...     # To make pytest.mark.skip work, we add the fake module 'mark', then set skip to `mock_decorator`
        ...     pytest.add_mock_module('mark')
        ...     pytest.mark.skip = mock_decorator
        ...
    
    Since we added the mock module ``mark``, and set the attribute ``skip`` to point at ``mock_decorator``, the
    test function ``test_something`` won't cause a syntax error. ``mock_decorator`` will just call test_something()
    which doesn't do anything anyway::
     
        >>> @pytest.mark.skip(reason="this test doesn't actually do anything...")
        ... def test_something():
        ...     pass
        >>>
        >>> def test_other_thing():
        ...     if True:
        ...         return pytest.skip('cannot test test_other_thing because of an error')
        ...
        >>>
    
    **Generating "disguised" mock classes**
    
    If you need the mock class to appear to have a certain class name and/or module path, you can generate
    "disguised" mock classes using :py:meth:`.make_mock_class` like so:
    
        >>> redis = Mocker.make_mock_class('Redis', module='redis')
        >>> redis
        <redis.Redis object at 0x7fd7402ea4a8>
    
    **A :class:`.Mocker` instance has the following behaviour**
    
    * Attributes that don't exist result in a function being returned, which accepts any arguments / keyword args,
      and simply returns ``None``
    
    Example::
    
        >>> m = Mocker()
        >>> repr(m.randomattr('hello', world=123))
        'None'
    
    
    * Arbitrary attributes ``x.something`` and items ``x['something']`` can be set on an instance, and they will
      be similarly returned when they're accessed. Attributes and items share the same key/value's, so the
      following examples are all accessing the same data::
    
    Example::
    
        >>> m = Mocker()
        >>> m.example = 'hello'
        >>> m['example'] = 'world'
        >>> print(m.example)
        world
        >>> print(m['example'])
        world
    
    * You can add arbitrary "modules" to a Mocker instance. With only the ``name`` argument, :py:meth:`.add_mock_module`
      will add a "module" under the instance, which is really just another :class:`.Mocker` instance.
    
    Example::
    
        >>> m = Mocker()
        >>> m.add_mock_module('my_module')
        >>> m.my_module.example = 'hello'
        >>> print(m.my_module['example'], m.my_module.example)
        hello hello
    
    
    """
    mock_modules: dict
    mock_attrs: dict
    
    def __init__(self, modules: dict = None, attributes: dict = None):
        self.mock_attrs = {} if attributes is None else attributes
        self.mock_modules = {} if modules is None else modules
    
    @classmethod
    def make_mock_class(cls, name='Mocker', instance=True, **kwargs):
        """
        Return a customized mock class or create an instance which appears to be named ``name``
        
        Allows code which might check ``x.__class__.__name__`` to believe it's the correct object.
        
        Using the kwarg ``module`` you can change the module that the class / instance appears to have been imported
        from, allowing for quite deceiving fake classes and instances.
        
        **Example usage**::
        
            >>> redis = Mocker.make_mock_class('Redis', module='redis')
            >>> # As seen below, the class appears to be called Redis, and even claims to be from the module `redis`
            >>> redis
            <redis.Redis object at 0x7fd7402ea4a8>
            >>> print(f'Module: {redis.__module__} - Class Name: {redis.__class__.__name__}')
            Module: redis - Class Name: Redis
        
        **Creating methods/attributes dynamically**
        
        You can set arbitrary attributes to point at a function, or just set them to a lambda::
        
            >>> redis.exists = lambda key: 1
            >>> redis.exists('hello')
            1
            >>> redis.hello()  # Non-existent attributes just act as a function that eats any args and returns None
            None
            
        
        :param name: The name to write onto the mock class's ``__name__`` (and ``__qualname__`` if not specified)
        :param bool instance: If ``True`` then the disguised mock class will be returned as an instance. Otherwise
                              the raw class itself will be returned for you to instantiate yourself.
        :param kwargs: All kwargs (other than ``qualname``) are forwarded to ``__init__`` of the disguised class
                       if ``instance`` is True.
        :key str qualname: Optionally specify the "qualified name" to insert into ``__qualname__``. If this isn't
                           specified, then ``name`` is used for qualname, which is fine for most cases anyway.
        :key str module: Optionally override the module namespace that the class is supposedly from. If not specified,
                         then the class will just inherit this module (``privex.helpers.common``)
        :return:
        """
        qualname = kwargs.pop('qualname', name)
        
        class OuterMocker(cls):
            pass
        
        OuterMocker.__name__ = name
        OuterMocker.__qualname__ = qualname
        
        if 'module' in kwargs:
            OuterMocker.__module__ = kwargs['module']
        
        return OuterMocker() if instance else OuterMocker
        
    def add_mock_module(self, name: str, value=None, mock_attrs: dict = None, mock_modules: dict = None):
        """
        Add a fake sub-module to this Mocker instance.
        
        Example::
        
            >>> m = Mocker()
            >>> m.add_mock_module('my_module')
            >>> m.my_module.example = 'hello'
            >>> print(m.my_module['example'], m.my_module.example)
            hello hello
        
        
        :param str name: The name of the module to add.
        :param value: Set the "module" to this object, instead of an instance of :class:`.Mocker`
        :param dict mock_attrs: If ``value`` is ``None``, then this can optionally contain a dictionary of
                                attributes/items to pre-set on the Mocker instance.
        :param dict mock_modules: If ``value`` is ``None``, then this can optionally contain a dictionary of
                                 "modules" to pre-set on the Mocker instance.
        """
        mock_attrs = {} if mock_attrs is None else mock_attrs
        mock_modules = {} if mock_modules is None else mock_modules
        
        self.mock_modules[name] = Mocker(modules=mock_modules, attributes=mock_attrs) if value is None else value

    def __getattribute__(self, item):
        try:
            return super().__getattribute__(item)
        except AttributeError:
            pass
        try:
            if item in super().__getattribute__('mock_modules'):
                return self.mock_modules[item]
        except AttributeError:
            pass
        try:
            if item in super().__getattribute__('mock_attrs'):
                return self.mock_attrs[item]
        except AttributeError:
            pass
        
        return lambda *args, **kwargs: None
    
    def __setattr__(self, key, value):
        if key in ['mock_attrs', 'mock_modules']:
            return super().__setattr__(key, value)
        m = super().__getattribute__('mock_attrs')
        m[key] = value
    
    def __getitem__(self, item):
        return self.__getattribute__(item)
    
    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    @property
    def __name__(self):
        return self.__class__.__name__






import binascii
import base64
import logging
from io import TextIOWrapper
from typing import Optional, Union, Tuple, Type
# characters that shouldn't be mistaken
from cryptography.exceptions import InvalidSignature
from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf import KeyDerivationFunction
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt

from privex.helpers.crypto.base import is_base64
from privex.helpers.exceptions import EncryptKeyMissing, EncryptionError
from privex.helpers.common import empty, random_str, ALPHANUM

log = logging.getLogger(__name__)


class EncryptHelper:
    """
    Symmetric AES-128 encryption/decryption made painless - wrapper class for :py:class:`cryptography.fernet.Fernet`

    A wrapper class for the :py:class:`cryptography.fernet.Fernet` encryption system, designed to make usage of Fernet
    as painless as possible.

    The class :class:`.EncryptHelper` contains various methods for simplifying the use of the Python library
    :doc:`Cryptography <cryptography:index>` 's :py:class:`cryptography.fernet.Fernet` system.

    :py:meth:`.encrypt_str` / :py:meth:`.decrypt_str` facilitate painless encryption and decryption of data using AES-128 CBC.
    They can either be passed a 32-byte Fernet key (base64 encoded) as an argument, or leave the key as None and they'll try
    to use the key defined on the :class:`.EncryptHelper` instance at :py:attr:`.EncryptHelper.encrypt_key`

    **Basic usage:**

        >>> from privex.helpers import EncryptHelper
        >>> key = EncryptHelper.generate_key()     # Generates a 32-byte symmetric key, returned as a base64 encoded string
        >>> crypt = EncryptHelper(key)             # Create an instance of EncryptHelper, en/decrypting using ``key`` by default
        # Encrypts the string 'hello world' with AES-128 CBC using the instance's key, returned as a base64 string
        >>> enc = crypt.encrypt_str('hello world')
        >>> print(enc)
        gAAAAABc7ERTpu2D_uven3l-KtU_ewUC8YWKqXEbLEKrPKrKWT138MNq-I9RRtCD8UZLdQrcdM_IhUU6r8T16lQkoJZ-I7N39g==

        >>> crypt.is_encrypted(enc)       # Check if a string/bytes is encrypted (only works with data matching the key)
        True
        >>> data = crypt.decrypt_str(enc) # Decrypt the encrypted data using the same key, outputs as a string
        >>> print(data)
        hello world


    """
    encrypt_key: str
    """A base64 encoded :class:`.Fernet` key, used by default for functions such as :py:meth:`.encrypt_str`"""
    
    def __init__(self, encrypt_key: str, **kwargs):
        """
        Create an instance of :class:`.EncryptHelper` using the :py:class:`cryptography.fernet.Fernet` key ``encrypt_key``
        as the default key for encrypting/decrypting data.

        :param str encrypt_key: Base64 encoded Fernet key, used by default for encrypting/decrypting data
        """
        self.settings = {**dict(encrypt_key=encrypt_key), **kwargs}
        self.encrypt_key = encrypt_key
    
    @staticmethod
    def generate_key(output: Optional[Union[str, TextIOWrapper]] = None, mode='w') -> str:
        """
        Generate a compatible encryption key for use with :py:class:`cryptography.fernet.Fernet`

        **NOTE:** Regardless of whether or not the method is outputting the key to a filename / stream, this
        method will always return the encryption key as a string after completion. The key returning was redacted
        from the outputting examples to help readability.

        **Examples**

        With no arguments, it will simply return the key as a string.

            >>> EncryptHelper.generate_key()
            '6vJ_o8XQRmX_TgUFTWWV_U2vm71ThnpWsCIvgXFWg9s='

        If ``output`` is a ``str`` - it's assumed to be a filename, and the Fernet key will be outputted
        to the file ``output`` using ``open(output, mode)`` (where ``mode`` defaults to ``'w'``).

        Below, we call ``generate_key`` with the string **test.key.txt** - and we can then see the file was created
        and contains the Fernet key encoded with Base64.

            >>> EncryptHelper.generate_key('test.key.txt')
            >>> open('test.key.txt').read()
            'aRDR-gCrmrPrMr9hQnL4epIPl2Szbzfid_vSTO-rl20='

        If ``output`` is a file/stream object, the method ``output.write(key)`` will be called, where ``key``
        is the Fernet key as a string.

        Below, we open **test2.key.txt** in write mode manually, then pass the file stream object to generate_key, which
        writes the key to the file.

            >>> with open('test2.key.txt', 'w') as fp:
            ...     EncryptHelper.generate_key(fp)
            >>> open('test2.key.txt').read()
            'DAFEvRkwG7ws0ccjIv2QL_s5cpeWktqpbc7eSjL-V74='


        :param None output: Simply return the generated key
        :param str output: Output the generated key to the filename ``output`` using the open mode ``mode``
        :param TextIOWrapper output: Output the generated key to the file/stream object ``output`` using ``.write(key: str)``

        :param str mode: If you're passing a string filename as ``output`` - then this controls the ``open()`` mode, e.g. 'w', 'a', 'w+'

        :return str key: The generated Fernet key, encoded with Base64.
        """
        key = Fernet.generate_key().decode('utf-8')
        if type(output) is str:
            log.debug('Outputting Fernet key to file "%s" using open mode "%s"', output, mode)
            with open(output, mode) as fp:
                fp.write(key)
        elif isinstance(output, TextIOWrapper):
            log.debug('Outputting Fernet key to file stream object: "%s"', repr(output))
            output.write(key)
        return key
    
    @staticmethod
    def password_key(password, salt=None, kdf: Type[KeyDerivationFunction] = PBKDF2HMAC, **kwargs) -> Tuple[str, dict]:
        """
        Generate a :py:class:`cryptography.fernet.Fernet` key based on a password and salt. Key derivation is customisable.

        :param password: A password to generate the key from, as ``str`` or ``bytes``
        :param     salt: The salt to use when generating the key, as ``str`` or ``bytes``
                         If ``salt`` is a string, it can also be passed in base64 format.


        **Standard Usage with manual salt:**

        Call ``password_key`` with a password of your choice, and a salt of your choice (ideally at least 16 chars),
        and a tuple containing the Fernet key (base64 encoded), and key derivative configuration will be returned.

            >>> ek = EncryptHelper
            >>> key, kd = ek.password_key('MySecurePass', salt=b'Sup3rseCr3tsalt')
            >>> key
            'rJ_g-lBT7pxeu4MVrhfi5rAv9yLbX5pTm6vkJj_Mezc='
            >>> kd
            {'length': 32, 'salt': 'U3VwM3JzZUNyM3RzYWx0', 'backend': <...Backend object at 0x7fd1c0220eb8>,
            'algorithm': <...SHA256 object at 0x7fd1b0232278>, 'iterations': 100000, 'kdf': <class '...PBKDF2HMAC'>}

        You can see when we call the method a second time with the same password and salt, we get the same Fernet key.

            >>> key, kd = ek.password_key('MySecurePass', salt=b'Sup3rseCr3tsalt')
            >>> key
            'rJ_g-lBT7pxeu4MVrhfi5rAv9yLbX5pTm6vkJj_Mezc='

        Now we can simply initialise the class with this key, and start encrypting/decrypting data:

            >>> enc = EncryptHelper(key)
            >>> mydata = enc.encrypt_str('hello')
            >>> mydata
            'gAAAAABdsJrpZvQAhAEwAGk2GPeJMUjUdp1FHAg42ncArvvQjqGztLslgexF7dKWbJ8bhYNt9MBzzT0WR_XEvl1j5Q95UOVTsQ=='
            >>> enc.decrypt_str(mydata)
            'hello'

        **Automatic salt generation:**

        While it's strongly recommend that you pass your own ``salt`` (at least 16 bytes recommended), for convenience this
        method will automatically generate a 16 byte salt and return it as part of the dict (second tuple item) returned.

        First, we generate a key from the password ``helloworld``

            >>> ek = EncryptHelper
            >>> key, kd = ek.password_key('helloworld')
            >>> key
            '6asAQ0qTQtmjw54RBR_RVmwsyv6EgTY_lcnVgJAVKCQ='
            >>> kd
            {'length': 32, 'salt': 'bDU5MzJaaEhnZ1htSmlQeg==', 'backend': <...Backend object at 0x7ff968053860>,
             'algorithm': <...SHA256 object at 0x7ff9685f6160>, 'iterations': 100000, 'kdf': <class '...PBKDF2HMAC'>}

        If we call **password_key** again with ``helloworld``, you'll notice it outputs a completely different key.
        This is because no salt was specified, so it simply generated yet another salt.

            >>> ek.password_key('helloworld')[0]
            'BfesIzfEPodtHSyPrpnkK0iDipHikaE7T1uuFFPnqmc='

        To actually get the same Fernet key back, we have to either:

         * Pass the entire ``kd`` dictionary as kwargs (safest option, contains all params used for generation)

            >>> ek.password_key('helloworld', **kd)[0]
            '6asAQ0qTQtmjw54RBR_RVmwsyv6EgTY_lcnVgJAVKCQ='

         * Pass the generated salt from the ``kd`` object, alongside our password.

            >>> ek.password_key('helloworld', salt=kd['salt'])[0]
            '6asAQ0qTQtmjw54RBR_RVmwsyv6EgTY_lcnVgJAVKCQ='


        """
        had_salt = salt is not None
        # If we were passed a salt, and it was a string, then check if it appears to be Base64
        # If the salt was Base64, then decode it into bytes.
        salt = base64.urlsafe_b64decode(salt) if had_salt and type(salt) is str and is_base64(salt) else salt
        # Generate a salt if one wasn't passed, and ensure that ``salt`` is bytes
        salt = random_str(16, chars=ALPHANUM) if not had_salt else salt
        salt = bytes(salt, 'utf-8') if type(salt) is not bytes else salt
        
        password = bytes(password, 'utf-8') if type(password) is not bytes else password
        
        # Default kwargs to be passed to the key deriv function if user doesn't override them.
        defaults = dict(length=32, salt=salt, backend=default_backend())
        
        if kdf is PBKDF2HMAC:
            defaults = {**defaults, **dict(algorithm=hashes.SHA256(), iterations=100000)}
        if kdf is Scrypt:
            defaults = {**defaults, **dict(n=2 ** 20, r=8, p=1)}
        # Merge our defaults with the users kwargs, initialise the KDF, then derive a Fernet key from the password
        kdf_args = {**defaults, **kwargs}
        k = kdf(**kdf_args)
        key = base64.urlsafe_b64encode(k.derive(password)).decode('utf-8')
        # To assist with future runs of this function, we return a dict containing the name of the KDF class used,
        # the salt (base64 encoded for easy handling), plus all of the KDF kwargs including our defaults
        kdf_args['salt'] = base64.urlsafe_b64encode(salt).decode('utf-8')
        kdf_args['kdf'] = kdf
        return key, kdf_args
    
    @classmethod
    def from_password(cls, password: Union[str, bytes], salt: Union[str, bytes], **settings):
        """
        Create an instance of :class:`.EncryptHelper` (or inheriting class) from a password derived
        Fernet key, instead of a pre-generated Fernet key.

        See :py:meth:`.password_key` for more detailed usage information.

        **Example**

            >>> enc = EncryptHelper.from_password('MySecurePass', salt=b'Sup3rseCr3tsalt')
            >>> d = enc.encrypt('hello')
            >>> enc.decrypt(d)
            'hello'


        :param password: A password to generate the key from, as ``str`` or ``bytes``
        :param     salt: The salt to use when generating the key, as ``str`` or ``bytes``
                         If ``salt`` is a string, it can also be passed in base64 format.

        """
        key, _ = cls.password_key(password=password, salt=salt, **settings)
        return cls(encrypt_key=key)
    
    @classmethod
    def from_file(cls, obj: Union[str, TextIOWrapper], **settings):
        """
        Create an instance of :class:`.EncryptHelper` (or inheriting class) using a Fernet key loaded from a file,
        or stream object.

            >>> enc = EncryptHelper.from_file('/home/john/fernet.key')
            >>> d = enc.encrypt('hello')
            >>> enc.decrypt(d)
            'hello'

        :param str obj: Load the key from the filename ``obj``
        :param TextIOWrapper obj: Load the key from the file/stream object ``obj`` using ``.read()``
        """
        if isinstance(obj, TextIOWrapper):
            return cls(encrypt_key=obj.read(), **settings)
        with open(str(obj)) as fp:
            return cls(encrypt_key=fp.read(), **settings)
    
    def get_fernet(self, key: Union[str, bytes] = None) -> Fernet:
        """
        Used internally for getting Fernet instance with auto-fallback to :py:attr:`.encrypt_key` and exception handling.

        :param str key: Base64 Fernet symmetric key for en/decrypting data. If empty, will fallback to :py:attr:`.encrypt_key`
        :raises EncryptKeyMissing: Either no key was passed, or something is wrong with the key.
        :return Fernet f: Instance of Fernet using passed ``key`` or self.encrypt_key for encryption.
        """
        if empty(key) and empty(self.encrypt_key):
            raise EncryptKeyMissing('No key argument passed, and ENCRYPT_KEY is empty. Cannot encrypt/decrypt.')
        
        key = self.encrypt_key if empty(key) else key
        try:
            f = Fernet(key)
            return f
        except (binascii.Error, ValueError):
            raise EncryptKeyMissing('The passed ``key`` or self.encrypt_key is not a valid Fernet key')
    
    def is_encrypted(self, data: Union[str, bytes], key: Union[str, bytes] = None) -> bool:
        """
        Returns True if the passed ``data`` appears to be encrypted. Can only verify encryption if the same ``key``
        that was used to encrypt the data is passed.

        :param str data: The data to check for encryption, either as a string or bytes
        :param str key:  Base64 encoded Fernet symmetric key for decrypting data. If empty, fallback to :py:attr:`.encrypt_key`
        :raises EncryptKeyMissing: Either no key was passed, or something is wrong with the key.
        :return bool is_encrypted: True if the data is encrypted, False if it's not encrypted or wrong key used.
        """
        f = self.get_fernet(key)
        
        # Convert the passed data into bytes before trying to decode it
        data = str(data).encode('utf-8') if type(data) != bytes else data
        
        # Attempt to extract the Fernet timestamp from the passed data. If exceptions are raised, then it's not encrypted.
        try:
            ts = f.extract_timestamp(data)
            log.debug(f'data was encrypted, token timestamp is {ts}')
            return True
        except (InvalidSignature, InvalidToken, binascii.Error) as e:
            log.debug('data is not encrypted? exception was: %s %s', type(e), str(e))
            return False
    
    def _crypt_str(self, direction: str, data: Union[str, bytes], key: Union[str, bytes] = None) -> str:
        """
        Used internally by :py:meth:`.encrypt_str` and :py:meth:`.decrypt_str`

        :param str direction: Either 'encrypt' or 'decrypt'
        :param str data:      The data to encrypt or decrypt as either a string or bytes
        :param str key:       Base64 encoded Fernet symmetric key for encrypting/decrypting data.
        :return str data_out: Either the encrypted data as a base64 encoded string, or decrypted data as a plain string.
        """
        if direction not in ['encrypt', 'decrypt']:
            raise ValueError('_crypt_str direction must be "encrypt" or "decrypt"')
        
        f = self.get_fernet(key)
        
        # Handle encryption/decryption of ``data``
        try:
            # If ``data`` isn't already bytes, cast to a string and convert it to bytes before encrypting/decrypting
            data = str(data).encode('utf-8') if type(data) != bytes else data
            out = f.encrypt(data) if direction == 'encrypt' else f.decrypt(data)
            return out.decode()  # Return encrypted/decrypted data as a string, not bytes.
        except Exception:
            strdat = str(data) if type(data) != bytes else str(data.decode())
            log.exception(f'An exception occurred while trying to {direction} the data starting with "{strdat:.4}"...')
            raise EncryptionError(f'Failed to {direction} data... An admin must check the logs.')
    
    def encrypt_str(self, data: Union[str, bytes], key: Union[str, bytes] = None) -> str:
        """
        Encrypts a piece of data ``data`` passed as a string or bytes using Fernet with the passed 32-bit symmetric
        encryption key ``key``. Outputs the encrypted data as a Base64 string for easy storage.

        The ``key`` cannot just be a random "password", it must be a 32-byte key encoded with URL Safe base64. Use the
        method :py:meth:`.generate_key` to create a Fernet compatible encryption key.

        Under the hood, Fernet uses AES-128 CBC to encrypt the data, with PKCS7 padding and HMAC_SHA256 authentication.

        If the ``key`` parameter isn't passed, or is empty (None / ""), then it will attempt to fall back to
        ``self.encrypt_key`` - if that's also empty, EncryptKeyMissing will be raised.

        :param str data:  The data to be encrypted, in the form of either a str or bytes.
        :param str key:   A Fernet encryption key (base64) to be used, if left blank will fall back to :py:attr:`.encrypt_key`
        :raises EncryptKeyMissing: Either no key was passed, or something is wrong with the key.
        :raises EncryptionError:   Something went wrong while attempting to encrypt the data
        :return str encrypted_data:   The encrypted version of the passed ``data`` as a base64 encoded string.
        """
        
        return self._crypt_str('encrypt', data, key)
    
    def decrypt_str(self, data: Union[str, bytes], key: Union[str, bytes] = None) -> str:
        """
        Decrypts ``data`` previously encrypted using :py:meth:`.encrypt_str` with the same Fernet compatible ``key``, and
        returns the decrypted version as a string.

        The ``key`` cannot just be a random "password", it must be a 32-byte key encoded with URL Safe base64. Use the
        method :py:meth:`.generate_key` to create a Fernet compatible encryption key.

        Under the hood, Fernet uses AES-128 CBC to encrypt the data, with PKCS7 padding and HMAC_SHA256 authentication.

        If the ``key`` parameter isn't passed, or is empty (None / ""), then it will attempt to fall back to
        :py:attr:`.encrypt_key` - if that's also empty, EncryptKeyMissing will be raised.

        :param str data:  The base64 encoded data to be decrypted, in the form of either a str or bytes.
        :param str key:   A Fernet encryption key (base64) for decryption, if blank, will fall back to :py:attr:`.encrypt_key`
        :raises EncryptKeyMissing: Either no key was passed, or something is wrong with the key.
        :raises EncryptionError:   Something went wrong while attempting to decrypt the data
        :return str decrypted_data:   The decrypted data as a string
        """
        return self._crypt_str('decrypt', data, key)




"""
Functions and classes related to working with Python's native asyncio support

To avoid issues with the ``async`` keyword, this file is named ``asyncx`` instead of ``async``

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import asyncio


def run_sync(func, *args, **kwargs):
    """
    Run an async function synchronously (useful for REPL testing async functions)

    Usage:

        >>> async def my_async_func(a, b, x=None, y=None):
        ...     return a, b, x, y
        >>>
        >>> run_sync(my_async_func, 1, 2, x=3, y=4)
        (1, 2, 3, 4,)

    :param callable func: An asynchronous function to run
    :param args:          Positional arguments to pass to ``func``
    :param kwargs:        Keyword arguments to pass to ``func``

    """
    coro = asyncio.coroutine(func)
    future = coro(*args, **kwargs)
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(future)


def async_sync(f):
    """
    Async Synchronous Decorator, borrowed from https://stackoverflow.com/a/23036785/2648583 - added this PyDoc comment
    and support for returning data from a synchronous function

    Allows a non-async function to run async functions using ``yield from`` - and can also return data

    Useful for unit testing, since unittest.TestCase functions are synchronous.

    Example async function:

        >>> async def my_async_func(a, b, x=None, y=None):
        ...     return a, b, x, y
        ...

    Using the above async function with a non-async function:

        >>> @async_sync
        ... def sync_function():
        ...     result = yield from my_async_func(1, 2, x=3, y=4)
        ...     return result
        ...
        >>> r = sync_function()
        >>> print(r)
        (1, 2, 3, 4,)
        >>> print(r[1])
        2


    """

    def wrapper(*args, **kwargs):
        coro = asyncio.coroutine(f)
        future = coro(*args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(future)

    return wrapper



"""
Exception classes used either by our helpers, or just generic exception names which are missing from
the standard base exceptions in Python, and are commonly used across our projects.

**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

"""


class PrivexException(Exception):
    """Base exception for all custom Privex exceptions"""
    pass


#####
# Exceptions related to DNS (domain, or individual record errors)
#####

class BaseDNSException(PrivexException):
    """Base exception for DNS-related exceptions"""
    pass


class BoundaryException(BaseDNSException):
    """
    Thrown when the v4/v6 address boundary for a reverse DNS record is invalid.
    
    An "address boundary" is the minimum amount of bits that defines a character in a reverse DNS domain 
    (i.e. those ending in in-addr/ip6.arpa):

         - IPv6 uses "nibbles" of 4 bits for each individual character in a standard 128-bit address
         - IPv4 uses "octets" of 8 bits, which represent 0-255 in each of the 4 "blocks" within an IPv4 address
    
    This means the minimum and maximum boundaries for IPv4 and IPv6 are as such:

        IPv6 - Minimum boundary: 4 bits (``*.f.ip6.arpa``), Maximum boundary: 128 bits (rDNS for a single address, i.e. a /128)
        IPv4 - Minimum boundary: 8 bits (``x.x.x.127.in-addr.arpa``), Maximum boundary: 32 bits (``1.0.0.127.in-addr.arpa``)
    
    """
    pass


class DomainNotFound(BaseDNSException):
    """Thrown when a (sub)domain or it's parent(s) could not be found"""
    pass


class InvalidDNSRecord(BaseDNSException):
    """Thrown when a passed DNS record is not valid"""
    pass


class CacheNotFound(PrivexException):
    """
    Thrown when a cache key is requested, but it doesn't exist / is expired.
    
    Most likely only used when some form of "strict mode" is enabled for the cache adapter.
    """
    pass


class NotConfigured(PrivexException):
    """
    Thrown when code attempts to access something that wasn't fully configured / instantiated by the user.
    
    Example: Attempting to use a database dependant function/method without having configured any database details.
    """


class NetworkUnreachable(PrivexException):
    """
    Thrown when a network interface or IP version (e.g. IPv4/v6) is unavailable.
    
    Example: when running ``ping`` with an IPv6 address on a system which has no IPv6.
    """


class EncryptionError(PrivexException):
    """Raised when something went wrong attempting to encrypt or decrypt a piece of data"""
    pass


class EncryptKeyMissing(EncryptionError):
    """Raised when ENCRYPT_KEY is not set, or invalid"""
    pass


class InvalidFormat(EncryptionError):
    """
    Raised when an invalid public/private format, or encoding is specified when serializing an asymmetric key pair
    """





"""
Various small helper functions/classes often used by classes such as :py:mod:`.EncryptHelper` and :py:mod:`.KeyManager`

"""
import base64
from typing import Union

from privex.helpers import byteify


def is_base64(sb: Union[str, bytes], urlsafe: bool = True) -> bool:
    """
    Returns ``True`` if ``sb`` appears to be a Base64 encoded string/bytes, otherwise ``False``

    :param str|bytes sb: Data to check for base64 encoding
    :param bool urlsafe: (Default: ``True``) Use :py:func:`base64.urlsafe_b64encode` and decode, instead of
                         plain :py:func:`base64.b64encode` and decode.
    :return bool is_base64: ``True`` if data appears to be Base64, otherwise ``False``
    """
    b64_enc = base64.urlsafe_b64encode if urlsafe else base64.b64encode
    b64_dec = base64.urlsafe_b64decode if urlsafe else base64.b64decode
    try:
        if isinstance(sb, str):
            # If there's any unicode here, an exception will be thrown and the function will return false
            sb = bytes(sb, 'ascii')
        if not isinstance(sb, bytes):
            raise ValueError("Argument 'sb' must be string or bytes")
        return b64_enc(b64_dec(sb)) == sb
    except Exception:
        return False


def auto_b64decode(data: Union[str, bytes], urlsafe: bool = True) -> bytes:
    """
    Determines if ``data`` is base64 encoded. If it is, then it will decode it and return the decoded bytes.
    Otherwise, it will return the original data as bytes.
    
    **Example usage**
    
        >>> a = b'dGVzdGluZw=='
        >>> b = 'hello world'
        >>> auto_b64decode(a)  # This is detected to be Base64 encoded, and so the decoded data is returned
        b'testing'
        >>> auto_b64decode(b)  # This does not appear to be Base64, and so is returned as normal (but as bytes)
        b'hello world'
    
    :param str|bytes data: Data to check for base64, and decode if required
    :param bool urlsafe: (Default: ``True``) Use :py:func:`base64.urlsafe_b64decode`, instead of
                         plain :py:func:`base64.b64decode`
    :return bytes data: If ``data`` was base64, this will be the decoded bytes. Otherwise, the original data as bytes.
    """
    data = byteify(data)
    b64_dec = base64.urlsafe_b64decode if urlsafe else base64.b64decode
    if is_base64(data):
        return b64_dec(data)
    return data




#!/usr/bin/env python3
"""
    +===================================================+
    |                 © 2019 Privex Inc.                |
    |               https://www.privex.io               |
    +===================================================+
    |                                                   |
    |        Originally Developed by Privex Inc.        |
    |                                                   |
    |        Core Developer(s):                         |
    |                                                   |
    |          (+)  Chris (@someguy123) [Privex]        |
    |          (+)  Kale (@kryogenic) [Privex]          |
    |                                                   |
    +===================================================+

Copyright 2019     Privex Inc.   ( https://www.privex.io )

Permission is hereby granted, free of charge, to any person obtaining a copy of 
this software and associated documentation files (the "Software"), to deal in 
the Software without restriction, including without limitation the rights to use, 
copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
Software, and to permit persons to whom the Software is furnished to do so, 
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all 
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""
import warnings

from setuptools import setup, find_packages
from os.path import join, dirname, abspath
from privex.helpers import VERSION, extras_require

BASE_DIR = dirname(abspath(__file__))

with open(join(BASE_DIR, "README.md"), "r") as fh:
    long_description = fh.read()

extensions = ['full', 'cache', 'crypto', 'django', 'docs', 'dev', 'extras', 'net', 'tests', 'setuppy']

extra_commands = {}

try:
    # noinspection PyUnresolvedReferences
    import privex.helpers.setuppy.commands
    extra_commands['extras'] = privex.helpers.setuppy.commands.ExtrasCommand
    extra_commands['bump'] = privex.helpers.setuppy.commands.BumpCommand
except (ImportError, AttributeError) as e:
    warnings.warn('Failed to import privex.helpers.setuppy.commands - the commands "extras" and "bump" may not work.')
    warnings.warn(f'Error Reason: {type(e)} - {str(e)}')


setup(
    name='privex_helpers',

    version=VERSION,

    description='A variety of helper functions and classes, useful for many different projects',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Privex/python-helpers",
    author='Chris (Someguy123) @ Privex',
    author_email='chris@privex.io',

    license='MIT',
    install_requires=[
        'privex-loghelper>=1.0.4', 'python-dateutil',
    ],
    cmdclass=extra_commands,
    extras_require=extras_require(extensions),
    packages=find_packages(exclude=['tests', 'test.*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Configuration options for helpers, and services they depend on, such as Redis.

To override settings from your app::

    >>> from privex.helpers import settings
    >>> settings.REDIS_HOST = 'redis.example.org'
    >>> settings.REDIS_PORT = 1234


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
from os.path import dirname, abspath, join

BASE_DIR = dirname(dirname(dirname(abspath(__file__))))
"""The root folder of this project (i.e. where setup.py is)"""

VERSION_FILE = join(BASE_DIR, 'privex', 'helpers', '__init__.py')
"""The file containing the package version (for :py:func:`.setuppy.bump_version`)"""

REDIS_HOST = 'localhost'
"""Hostname / IP address where redis-server is running on"""

REDIS_PORT = 6379
"""Port number that Redis is running on at ``REDIS_HOST``"""

REDIS_DB = 0
"""Redis database to use (number)"""

DEFAULT_CACHE_TIMEOUT = 300
"""Default cache timeout in seconds, used by cache adapters in the module :py:mod:`.cache`"""

EXTRAS_FOLDER = 'extras'
"""Folder where additional requirements files can be found for :py:func:`privex.helpers.setuppy.common.extras`"""




"""
This module handles connection objects for databases, APIs etc. by exposing functions which initialise and
store class instances for re-use.

It's primarily intended to be used to enable database, caching and third-party API connectivity for the helpers in this
package, however, you're free to use the functions / classes / attributes exposed in this module for your own apps.

Classes are generally initialised using the settings from :py:mod:`.settings` - see the docs for that module to learn
how to override the settings if the defaults don't work for you.


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |        License: X11 / MIT                         |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )


"""
import logging
from privex.helpers import settings

log = logging.getLogger(__name__)

__all__ = [
    'HAS_REDIS', 'HAS_DNSPYTHON', 'HAS_CRYPTO', 'HAS_SETUPPY_BUMP', 'HAS_SETUPPY_COMMANDS', 'HAS_SETUPPY_COMMON'
]

HAS_REDIS = False
"""If the ``redis`` module was imported successfully, this will change to True."""

HAS_DNSPYTHON = False
"""If the ``dns.resolver`` module was imported successfully, this will change to True."""

HAS_CRYPTO = False
"""If :py:mod:`privex.helpers.crypto` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMON = False
"""If :py:mod:`privex.helpers.setuppy.common` was imported successfully, this will change to True"""

HAS_SETUPPY_BUMP = False
"""If :py:mod:`privex.helpers.setuppy.bump` was imported successfully, this will change to True"""

HAS_SETUPPY_COMMANDS = False
"""If :py:mod:`privex.helpers.setuppy.commands` was imported successfully, this will change to True"""

__STORE = {}
"""This ``dict`` is used to store initialised classes for connections to databases, APIs etc."""


try:
    import redis

    def get_redis() -> redis.Redis:
        """Get a Redis connection object. Create one if it doesn't exist."""
        if 'redis' not in __STORE:
            __STORE['redis'] = redis.Redis(host=settings.REDIS_HOST, port=settings.REDIS_PORT, db=settings.REDIS_DB)
        return __STORE['redis']
    
    def reset_redis() -> redis.Redis:
        """Close the global Redis connection, delete the instance, then re-instantiate it"""
        if 'redis' in __STORE:
            __STORE['redis'].close()
            del __STORE['redis']
        return get_redis()
    
    def configure_redis(host=settings.REDIS_HOST, port: int = settings.REDIS_PORT, db: int = settings.REDIS_DB):
        """Update global Redis settings and re-instantiate the global Redis instance with the new settings."""
        settings.REDIS_DB = db
        settings.REDIS_PORT = port
        settings.REDIS_HOST = host
        if 'redis' in __STORE:
            del __STORE['redis']
        return get_redis()


    __all__ += ['get_redis', 'reset_redis', 'configure_redis']
    HAS_REDIS = True

except ImportError:
    log.debug('privex.helpers __init__ failed to import "redis", Redis dependent helpers will be disabled.')
except Exception:
    log.debug('privex.helpers __init__ failed to import "redis", (unknown exception), '
              'Redis dependent helpers will be disabled.')




"""
Privex's Python Helpers - https://github.com/privex/python-helpers

X11 / MIT License

**Submodules**:

  * :py:mod:`.common` - Uncategorized functions and classes, including bool checks and data parsing
  * :py:mod:`.decorators` - Class / function decorators
  * :py:mod:`.django` - Django-specific functions/classes, only available if Django package is installed
  * :py:mod:`.net` - Network related functions/classes such as ASN name lookup, and IP version bool checks
  * :py:mod:`.exceptions` - Exception classes used either by our helpers, or generic exceptions for use in projects


**Copyright**::

        +===================================================+
        |                 © 2019 Privex Inc.                |
        |               https://www.privex.io               |
        +===================================================+
        |                                                   |
        |        Originally Developed by Privex Inc.        |
        |                                                   |
        |        Core Developer(s):                         |
        |                                                   |
        |          (+)  Chris (@someguy123) [Privex]        |
        |          (+)  Kale (@kryogenic) [Privex]          |
        |                                                   |
        +===================================================+

    Copyright 2019     Privex Inc.   ( https://www.privex.io )

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in 
    the Software without restriction, including without limitation the rights to use, 
    copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the 
    Software, and to permit persons to whom the Software is furnished to do so, 
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION 
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""

import logging
import warnings

from privex.helpers.common import *
from privex.helpers.collections import *
from privex.helpers.decorators import *
from privex.helpers.net import *
from privex.helpers.exceptions import *
from privex.helpers.plugin import *
from privex.helpers.cache import CacheNotFound, CacheAdapter, CacheWrapper, MemoryCache, cached

from privex.helpers import plugin

log = logging.getLogger(__name__)


try:
    from privex.helpers.cache.RedisCache import RedisCache
    plugin.HAS_REDIS = True
except ImportError:
    log.debug('privex.helpers __init__ failed to import "RedisCache", not loading RedisCache')

try:
    from privex.helpers.asyncx import async_sync, run_sync
except ImportError:
    log.debug('privex.helpers __init__ failed to import "asyncx", not loading async helpers')

try:
    from privex.helpers.crypto import Format, KeyManager, EncryptHelper, auto_b64decode, is_base64
    plugin.HAS_CRYPTO = True
except ImportError:
    log.debug('privex.helpers __init__ failed to import "crypto", not loading cryptography helpers')

try:
    from privex.helpers.setuppy.common import extras_require, reqs
    plugin.HAS_SETUPPY_COMMON = True
except ImportError:
    log.debug('privex.helpers __init__ failed to import "setuppy.common", not loading packaging helpers')

try:
    from privex.helpers.setuppy.commands import BumpCommand, ExtrasCommand
    plugin.HAS_SETUPPY_COMMANDS = True
except ImportError:
    log.debug('privex.helpers __init__ failed to import "setuppy.commands", not loading packaging helpers')

try:
    from privex.helpers.setuppy.bump import bump_version, get_current_ver
    plugin.HAS_SETUPPY_BUMP = True
except ImportError:
    log.debug('privex.helpers __init__ failed to import "setuppy.bump", not loading packaging helpers')

try:
    from privex.helpers.extras import *
except ImportError:
    log.debug('privex.helpers __init__ failed to import "extras", cannot load any extras.* modules...')

try:
    from privex.helpers.converters import *
except ImportError:
    log.debug('privex.helpers __init__ failed to import "converter", cannot load any converter.* modules...')


def _setup_logging(level=logging.WARNING):
    """
    Set up logging for the entire module ``privex.helpers`` . Since this is a package, we don't add any
    console or file logging handlers, we purely just set our minimum logging level to WARNING to avoid
    spamming the logs of any application importing it.
    """
    try:
        from privex.loghelper import LogHelper
        lh = LogHelper(__name__, level=level)
        return lh.get_logger()
    except ImportError:
        warnings.warn(f'{__name__} failed to import privex.loghelper. Logging may not work as expected.')
        lh = logging.getLogger(__name__)
        lh.setLevel(logging.WARNING)
        return log
    

log = _setup_logging()
name = 'helpers'

VERSION = '2.5.0'






"""
Requirements / distutils "extras" helpers


Example usage for requirements/extras helpers
---------------------------------------------

Examples::

    >>> extensions = ['cache', 'crypto', 'django']
    # Loads ``extras/{extension}.txt`` for each extension, then returns a dictionary mapping each extension
    # to their requirements
    >>> extras_require(extensions)
    {'cache': ['redis'], 'crypto': ['cryptography>=2.8'], 'django': ['Django']}

    # Load ``example.txt`` and merge sub-requirement lists.
    >>> reqs('example.txt')
    ['privex-loghelper>=1.0', 'privex-jsonrpc', 'requests']



License note
------------

The following functions were taken from Celery's ``setup.py`` file. Since the time of writing
this comment block, they may have been further modified from their originals.

 * strip_comments
 * pip_requirement
 * _reqs
 * reqs
 * extras

Celery is licensed under The BSD License (3 Clause, also known as the new BSD license). The license is an
OSI approved Open Source license and is GPL-compatible(1).

The BSD license text can also be found here: http://www.opensource.org/licenses/BSD-3-Clause

Celery LICENSE file: https://github.com/celery/celery/blob/master/LICENSE

"""
import logging
import os
from privex.helpers.settings import EXTRAS_FOLDER
from typing import Union, Dict

log = logging.getLogger(__name__)


def strip_comments(l: str) -> str:
    """
    Strip any ``#`` comments from a line.
    
    :param str l: A string line, which may contain ``#`` comments
    :return str clean_line: The line ``l`` - stripped of any comments and excess whitespace.
    """
    return l.split('#', 1)[0].strip()


def pip_requirement(req: str) -> list:
    """Check a requirement line for imports ``-r some/path/requirements.txt`` and import them if found."""
    if req.startswith('-r '):
        _, path = req.split()
        return reqs(*path.split('/'))
    return [req]


def _reqs(*f: str) -> list:
    lines = (strip_comments(l) for l in open(os.path.join(os.getcwd(), *f)).readlines())
    return [pip_requirement(r) for r in lines if r]


def reqs(*f: str) -> list:
    """
    Parse requirement file.
    
    Example:
        reqs('default.txt')          # $PWD/default.txt
        reqs('extras', 'redis.txt')  # $PWD/extras/redis.txt
    Returns:
        List[str]: list of requirements specified in the file.
    """
    return [req for subreq in _reqs(*f) for req in subreq]


def extras(*p) -> list:
    """Parse requirement in the extras/ directory."""
    return reqs(EXTRAS_FOLDER, *p)


def extras_require(extra: Union[list, set, tuple]) -> Dict[str, list]:
    """Get map of all extra requirements."""
    return {x: extras(x + '.txt') for x in extra}
